*** lib/kernel/src/dist_ac.erl	Mon Jan 28 09:41:44 2002
--- ../otp_src_R8B-1.fixed/lib/kernel/src/dist_ac.erl	Tue May  7 01:05:23 2002
***************
*** 25,26 ****
--- 25,27 ----
  	 takeover_application/2, permit_application/2, permit_only_loaded_application/2]).
+ -export([whereis_application/1]).
  
***************
*** 29,31 ****
  %% Internal exports
! -export([init/1, handle_cast/2, handle_call/3, handle_info/2, terminate/2,
  	 send_timeout/3]).
--- 30,32 ----
  %% Internal exports
! -export([init/1, handle_cast/2, handle_call/3, handle_info/2, terminate/2, code_change/3,
  	 send_timeout/3]).
***************
*** 38,43 ****
  -define(DIST_AC, ?MODULE).
! -define(LOCK_ID, ?MODULE).
  
  %% This is the protocol version for the dist_ac protcol (between nodes)
! -define(vsn, 1).
  
--- 39,48 ----
  -define(DIST_AC, ?MODULE).
! 
! -define(SERVER_LOCK_ID, ?MODULE).
! -define(USER_LOCK_ID, dist_ac_user).
! 
! -define(RND_TIME(T), random:uniform(T) ).
  
  %% This is the protocol version for the dist_ac protcol (between nodes)
! -define(vsn, 2).
  
***************
*** 117,118 ****
--- 122,124 ----
  %%-----------------------------------------------------------------
+ 
  load_application(AppName, DistNodes) ->
***************
*** 126,128 ****
  	    global:trans(
! 	      {?LOCK_ID, self()},
  	      fun() ->
--- 132,134 ----
  	    global:trans(
! 	      {?USER_LOCK_ID, self()},
  	      fun() ->
***************
*** 138,139 ****
--- 144,148 ----
  
+ whereis_application(AppName) ->
+     gen_server:call(?DIST_AC, {whereis_application, AppName}, infinity).
+ 
  %%-----------------------------------------------------------------
***************
*** 161,163 ****
      Nodes = get_nodes(AppName),
!     LockId = {?LOCK_ID, self()},
      global:trans(
--- 170,172 ----
      Nodes = get_nodes(AppName),
!     LockId = {?USER_LOCK_ID, self()},
      global:trans(
***************
*** 173,175 ****
      Nodes = get_nodes(AppName),
!     LockId = {?LOCK_ID, self()},
      global:trans(
--- 182,184 ----
      Nodes = get_nodes(AppName),
!     LockId = {?USER_LOCK_ID, self()},
      global:trans(
***************
*** 193,194 ****
--- 202,205 ----
      process_flag(trap_exit, true),
+     {MS,S,US} = now(),
+     random:seed(MS,S,US + erlang:phash(node(), 100000) ),
      {ok, #state{}}.
***************
*** 196,198 ****
  sync_dacs(Appls) ->
!     Res = global:trans({?LOCK_ID, sync_dacs},
  		       fun() ->
--- 207,209 ----
  sync_dacs(Appls) ->
!     Res = global:trans({?SERVER_LOCK_ID, sync_dacs},
  		       fun() ->
***************
*** 226,228 ****
  	%% don't handle version mismatch here.
! 	{dist_ac_new_node, Vsn, Node, HisAppls, HisStarted} ->
  	    monitor_node(Node, false),
--- 237,261 ----
  	%% don't handle version mismatch here.
! 
! 	%%
! 	%% Version 1 is useless, but there is no way to inform the remote node 
! 	%% about a version mismatch.
! 	%% If this happens just drop dead.
! 	%%
! 	{dist_ac_new_node, Vsn, OtherNode, _HisAppls, _HisStarted} when Vsn == 1 ->
! 	    %% Version mismatch
! 	    dist_ac_version_mismatch(Vsn, OtherNode);
! 
!  	{dist_ac_new_node, Vsn, OtherNode, HisAppls, HisStarted} ->
!  	    MyStarted = zf(fun(Appl) when Appl#appl.id == local ->
!  				   {true, {node(), Appl#appl.name}};
!  			      (_) ->
!  				   false
!  			   end, Appls),
!  	    {?DIST_AC, OtherNode} ! {dist_ac_new_node_resp, ?vsn, node(), Appls, MyStarted},
!  	    NAppls = dist_merge(Appls, HisAppls, OtherNode),
!  	    NRStarted = RStarted ++ HisStarted,
!  	    monitor_node(Node, false),
!  	    wait_dacs([Node|Nodes], KnownNodes, NAppls, NRStarted);
!  	
! 	{dist_ac_new_node_resp, Vsn, Node, HisAppls, HisStarted} ->
  	    monitor_node(Node, false),
***************
*** 231,232 ****
--- 264,266 ----
  	    wait_dacs(Nodes, [Node | KnownNodes], NAppls, NRStarted);
+ 
  	{nodedown, Node} ->
***************
*** 297,298 ****
--- 331,339 ----
  
+ handle_call({whereis_application, AppName}, _From, S) ->
+     case lists:keysearch(AppName, #appl.name, S#state.appls) of
+ 	{value, Appl} when record(Appl,appl) ->
+ 	    {reply,{ok,Appl#appl.id},S};
+ 	_ ->
+ 	    {reply,{error, {not_running_distributed, AppName}}, S}
+     end;
  
***************
*** 391,393 ****
  
- 
  handle_info({ac_load_application_req, AppName}, S) ->
--- 432,433 ----
***************
*** 403,406 ****
      %% We must decide if we or another node should start the application
!     Lock = {?LOCK_ID, self()},
!     case global:set_lock(Lock, [node()], 1) of
  	true ->
--- 443,448 ----
      %% We must decide if we or another node should start the application
!     Lock = {?SERVER_LOCK_ID, self()},
!     %%Nodes = [node()],
!     Nodes = S#state.known,
!     case global:set_lock(Lock, Nodes, 1) of
  	true ->
***************
*** 416,418 ****
  	false ->
! 	    send_after(100, {ac_start_application_req, AppName}),
  	    {noreply, S}
--- 458,460 ----
  	false ->
! 	    send_after(?RND_TIME(100), {ac_start_application_req, AppName}),
  	    {noreply, S}
***************
*** 501,514 ****
  %%-----------------------------------------------------------------
- handle_info({dist_ac_new_node, Vsn, Node, HisAppls, []}, S) ->
-     Appls = S#state.appls,
-     MyStarted = zf(fun(Appl) when Appl#appl.id == local ->
- 			   {true, {node(), Appl#appl.name}};
- 		      (_) ->
- 			   false
- 		   end, Appls),
-     {?DIST_AC, Node} ! {dist_ac_new_node, ?vsn, node(), Appls, MyStarted},
-     NAppls = dist_merge(Appls, HisAppls, Node),
-     {noreply, S#state{appls = NAppls, known = [Node | S#state.known]}};
  
! handle_info({dist_ac_app_started, Node, Name, Res}, S) ->
      case {keysearch(Name, #appl.name, S#state.appls), lists:member(Name, S#state.started)} of
--- 543,558 ----
  %%-----------------------------------------------------------------
  
! handle_info({dist_ac_app_state, Node, AppName}=Msg, S) ->
!     NS = handle_dist_request(Msg,S),
!     {noreply,NS};
! 
! handle_info({dist_ac_new_node, Vsn, Node, HisAppls, []}=Msg, S) ->
!     NS = handle_dist_request(Msg,S),
!     {noreply,NS};
! 
! handle_info({dist_ac_weight, Name, Weight, Node}=Msg, S) ->
!     NS = handle_dist_request(Msg,S),
!     {noreply,NS};
! 
! handle_info({dist_ac_app_started, Node, Name, Res}=Msg, S) ->
      case {keysearch(Name, #appl.name, S#state.appls), lists:member(Name, S#state.started)} of
***************
*** 560,607 ****
  	{value, Appl} ->
! 	    NAppl = Appl#appl{id = undefined},
  	    NAppls = keyreplace(AppName, #appl.name, Appls, NAppl),
  	    RStarted = keydelete(AppName, 2, S#state.remote_started),
! 	    {noreply, S#state{appls = NAppls, remote_started = RStarted}}
!     end;
! 
! handle_info({dist_ac_weight, Name, Weight, Node}, S) ->
!     %% This means another node starts up, and will eventually take over
!     %% this appl.  We have a situation like: {Name, [{Node}, node()]}
!     %% Node sends us this msg, and we must respond.  It doesn't really
!     %% matter what we send him; but it must be a dist_ac_weight msg.
!     %% Another situation is {Name, [RNode, {node()}, Node]}.
!     %%
!     %% Yet another situation is that the node where Name was running crashed,
!     %% and Node has got the nodedown message, but we haven't.  In this case,
!     %% we must send a correct weight to Node. i.e. the same weight that
!     %% we'll send to him later, when we get the nodedown message.
!     case keysearch(Name, #appl.name, S#state.appls) of
! 	{value, Appl} -> 
! 	    Id = Appl#appl.id,
! 	    case Id of 
! 		run_waiting ->
! 		    {?DIST_AC, Node} ! {dist_ac_weight, Name, 0, node()},
! 		    {noreply, S};
! 		undefined -> 
! 		    {noreply, 
! 		     S#state{tmp_locals = [{Name, Weight, Node} |
! 					   S#state.tmp_locals]}};
! 		{takeover, _} -> 
! 		    {noreply, 
! 		     S#state{tmp_locals = [{Name, Weight, Node} |
! 					   S#state.tmp_locals]}};
! 		{failover, _} -> 
! 		    {noreply, 
! 		     S#state{tmp_locals = [{Name, Weight, Node} |
! 					   S#state.tmp_locals]}};
! 		_ ->
! 		    MyWeight = get_cached_weight(Name, S),
! 		    {?DIST_AC, Node} ! {dist_ac_weight, Name, MyWeight, node()},
! 		    NTWs = keyreplaceadd(Name, 1, S#state.tmp_weights,
! 					 {Name, MyWeight}),
! 		    {noreply,  S#state{tmp_weights = NTWs}}
! 	    end;
! 	_ ->
! 	    {noreply, 
! 	     S#state{tmp_locals = [{Name, Weight, Node} | S#state.tmp_locals]}}
      end;
--- 604,629 ----
  	{value, Appl} ->
! 	    AppNames = dist_get_runnable(S#state.appls),
! 	    NAppl = case Appl#appl.id of
! 			{distributed,Node} when Node /= node() ->
! 			    case lists:member(Appl#appl.name,AppNames) of
! 				true ->
! 				    Appl#appl{id = {failover, Node}};
! 				_ ->
! 				    Appl#appl{id = undefined}
! 			    end;
! 			_ ->
! 			    Appl#appl{id = undefined}
! 		    end,
  	    NAppls = keyreplace(AppName, #appl.name, Appls, NAppl),
  	    RStarted = keydelete(AppName, 2, S#state.remote_started),
! 	    NS = S#state{appls = NAppls, remote_started = RStarted},
! 	    NS2 = case NAppl#appl.id of
! 		      {failover,_} ->
! 			  restart_appl(AppName,NS);
! 		      _ ->
! 			  ac_not_running(AppName),
! 			  NS
! 		  end,
! 	    {noreply, NS2}
! 	    %%{noreply, S#state{appls = NAppls, remote_started = RStarted}}
      end;
***************
*** 681,683 ****
  	true when IsHisApp == true ->
! 	    case catch start_appl(AppName, NewS, req) of
  		{ok, NewS2, _}  ->
--- 703,705 ----
  	true when IsHisApp == true ->
! 	    case catch locked_start_appl(AppName, NewS, req) of
  		{ok, NewS2, _}  ->
***************
*** 708,709 ****
--- 730,734 ----
  
+ code_change(OldVsn,S,Extra) ->
+     {ok,S}.
+ 
  %%%-----------------------------------------------------------------
***************
*** 772,774 ****
  
! start_appl(AppName, S, Type) ->
      %% Get nodes, and check if App is loaded on all involved nodes.
--- 797,908 ----
  
! sync_appl_state(AppName,S) ->
!     Nodes = S#state.known -- [node()],
!     Msg = {dist_ac_app_state, node(), AppName},
!     send_msg(Msg,Nodes),
!     {ApplMaster,NS} = collect_appl_state_answers(Nodes, AppName, undefined, S),
!     case ApplMaster of
! 	{distributed_pending,PendingNodes} ->
! 	    NS2 = dist_process_requests(NS,?RND_TIME(200)),
! 	    sync_appl_state(AppName,NS2);
! 
! 	undefined ->
! 	    %% Replace the local application id definition
! 	    case lists:keysearch(AppName,#appl.name,Appls=NS#state.appls) of
! 		{value,Appl} ->
! 		    NAppl = Appl#appl{id=ApplMaster},
! 		    NAppls = keyreplaceadd(AppName, #appl.name, Appls, NAppl),
! 		    NS#state{appls=NAppls};
! 		_ ->
! 		    NS
! 	    end;
! 
! 	{distributed,Node} ->
! 	    %% Replace the local application id definition
! 	    case lists:keysearch(AppName,#appl.name,Appls=NS#state.appls) of
! 		{value,Appl} ->
! 		    NAppl = Appl#appl{id=ApplMaster},
! 		    NAppls = keyreplaceadd(AppName, #appl.name, Appls, NAppl),
! 		    NS#state{appls=NAppls};
! 		_ ->
! 		    NS
! 	    end;
! 
! 	_ ->
! 	    NS
!     end.
! 
! collect_appl_state_answers([], AppName, Master, S) ->
!     {Master,S};
! collect_appl_state_answers([Node|Nodes], AppName, Master, S) ->
!     monitor_node(Node,true),
!     receive
!  	{dist_ac_new_node, Vsn, OtherNode, HisAppls, HisStarted}=Msg ->
! 	    NS = handle_dist_request(Msg,S),
! 	    monitor_node(Node,false),
! 	    collect_appl_state_answers([Node|Nodes], AppName, Master, NS);
! 	    
! 	{dist_ac_app_state, OtherNode, AppName}=Msg ->
! 	    %% Special mode to hint equal level not to start ...
! 	    NS = handle_dist_request({Msg,true},S),
! 	    monitor_node(Node,false),
! 	    collect_appl_state_answers([Node|Nodes], AppName, Master, NS);
! 
! 	{dist_ac_app_state, OtherNode, OtherAppName}=Msg ->
! 	    NS = handle_dist_request(Msg,S),
! 	    monitor_node(Node,false),
! 	    collect_appl_state_answers([Node|Nodes], AppName, Master, NS);
! 
! 	{dist_ac_weight, Name, Weight, OtherNode}=Msg when OtherNode /= node() ->
! 	    NS = handle_dist_request(Msg,S),
! 	    monitor_node(Node,false),
! 	    collect_appl_state_answers([Node|Nodes], AppName, Master, NS);
! 	    
! 	{dist_ac_app_state_resp, Node, {AppName, AppId, _}}=Msg
! 	when AppId == local; AppId == run_waiting ->
! 	    monitor_node(Node,false),
! 	    collect_appl_state_answers(Nodes, AppName, {distributed, Node}, S);
! 
! 	{dist_ac_app_state_resp, Node, {AppName, _, true}}=Msg ->
! 	    NewMaster = case Master of
! 			    {distributed,_} ->
! 				Master;
! 			    {distributed_pending,PendingNodes} ->
! 				{distributed_pending,[Node|PendingNodes]};
! 			    _ ->
! 				{distributed_pending,[Node]}
! 			end,
! 	    monitor_node(Node,false),
! 	    collect_appl_state_answers(Nodes, AppName, NewMaster, S);
! 
! 	{dist_ac_app_state_resp, Node, {AppName, _, _}}=Msg ->
! 	    monitor_node(Node,false),
! 	    collect_appl_state_answers(Nodes, AppName, Master, S);
! 
! 	{nodedown, Node} ->
! 	    %% Ugly: send to ourselfs to cleanup state for this node later
! 	    self() ! {nodedown, Node},
! 	    collect_appl_state_answers(Nodes, AppName, Master, S)
!     end.
! 
! locked_start_appl(AppName,S, Type) ->
!     %% We must decide if we or another node should start this application
!     Lock = {?SERVER_LOCK_ID, self()},
!     Nodes = S#state.known,
!     case global:set_lock(Lock, Nodes, 1) of
! 	true ->
! 	    Res = start_appl(AppName, S, Type),
! 	    global:del_lock(Lock),
! 	    Res;
! 
! 	false ->
! 	    NS = dist_process_requests(S, ?RND_TIME(100)),
! 	    locked_start_appl(AppName,NS, Type)
!     end.
!     
! start_appl(AppName, IS, Type) ->
!     %% Now sync the application state with other nodes since we ignore
!     %% events on applications not started locally.
!     %% If we don't sync we will force a failover.
!     S = sync_appl_state(AppName,IS),
! 
      %% Get nodes, and check if App is loaded on all involved nodes.
***************
*** 785,786 ****
--- 919,921 ----
  	    {ok,S,false};
+ 
  	_ ->
***************
*** 806,809 ****
  
! start_distributed(Appl, Name, Nodes, PermittedNodes, S, Type) ->
!     case find_start_node(Nodes, PermittedNodes, Name, S) of
  	{ok, Node} when Node == node() ->
--- 941,945 ----
  
! start_distributed(Appl, Name, Nodes, PermittedNodes, IS, Type) ->
!     {SN,S} = find_start_node(Nodes, PermittedNodes, Name, IS),
!     case SN of
  	{ok, Node} when Node == node() ->
***************
*** 851,852 ****
--- 987,989 ----
  	    {{distributed, Node}, false};
+ 
  	{dist_ac_app_started, Node, Name, {error, R}} ->
***************
*** 855,866 ****
  	    {Appl#appl.id, false};
! 	{dist_ac_weight, Name, _Weigth, Node} ->
! 	    %% This is the situation: {Name, [RNode, {Node}, node()]}
! 	    %% and permit(false) is called on RNode, and we sent the
! 	    %% weigth first.  Node handled it in handle_info, and
! 	    %% now we must send him a weigth msg.  We can use any weigth;
! 	    %% he wins anyway.
! 	    monitor_node(Node, false),
! 	    {?DIST_AC, Node} !
! 		{dist_ac_weight, Name, get_cached_weight(Name, S), node()},
! 	    wait_dist_start(Node, Appl, Name, Nodes, PermittedNodes, S, Type);
  	{nodedown, Node} ->
--- 992,1010 ----
  	    {Appl#appl.id, false};
! 
!  	{dist_ac_new_node, OtherVsn, OtherNode, HisAppls, HisStarted}=Msg ->
! 	    NS = handle_dist_request(Msg,S),
! 	    monitor_node(Node,false),
! 	    wait_dist_start(Node, Appl, Name, Nodes, PermittedNodes, NS, Type);
! 
! 	{dist_ac_app_state, OtherNode, OtherAppName}=Msg ->
! 	    NS = handle_dist_request(Msg,S),
! 	    monitor_node(Node,false),
! 	    wait_dist_start(Node, Appl, Name, Nodes, PermittedNodes, NS, Type);
! 
! 	{dist_ac_weight, OtherName, OtherWeight, OtherNode}=Msg 
! 	when OtherNode /= node() ->
! 	    NS = handle_dist_request(Msg,S),
! 	    monitor_node(Node,false),
! 	    wait_dist_start(Node, Appl, Name, Nodes, PermittedNodes, NS, Type);
! 	    
  	{nodedown, Node} ->
***************
*** 883,884 ****
--- 1027,1029 ----
  	    {{distributed, Node}, false};
+ 
  	{dist_ac_app_started, Node, Name, {error, R}} ->
***************
*** 886,887 ****
--- 1031,1046 ----
  	    {Appl#appl.id, false};
+ 
+  	{dist_ac_new_node, OtherVsn, OtherNode, HisAppls, HisStarted}=Msg ->
+ 	    NS = handle_dist_request(Msg,S),
+ 	    wait_dist_start2(Appl, Name, Nodes, PermittedNodes, NS, Type);
+ 
+ 	{dist_ac_app_state, OtherNode, OtherAppName}=Msg ->
+ 	    NS = handle_dist_request(Msg,S),
+ 	    wait_dist_start2(Appl, Name, Nodes, PermittedNodes, NS, Type);
+ 
+ 	{dist_ac_weight, OtherName, OtherWeight, OtherNode}=Msg 
+ 	when OtherNode /= node() ->
+ 	    NS = handle_dist_request(Msg,S),
+ 	    wait_dist_start2(Appl, Name, Nodes, PermittedNodes, NS, Type);
+ 	    
  	{nodedown, Node} ->
***************
*** 944,945 ****
--- 1103,1127 ----
  
+ dist_process_requests(S,Timeout) ->
+     receive
+  	{dist_ac_new_node, Vsn, OtherNode, HisAppls, HisStarted}=Msg ->
+ 	    NS = handle_dist_request(Msg,S),
+ 	    dist_process_requests(NS,Timeout);
+ 	
+ 	{dist_ac_app_state, OtherNode, OtherAppName}=Msg ->
+ 	    NS = handle_dist_request(Msg,S),
+ 	    dist_process_requests(NS,Timeout);
+ 	
+ 	{dist_ac_weight, Name, Weight, OtherNode}=Msg when OtherNode /= node() ->
+ 	    NS = handle_dist_request(Msg,S),
+ 	    dist_process_requests(NS,Timeout);
+ 
+ 	{dist_ac_app_loaded, OtherNode, Name, HisNodes, Permission, HeKnowsMe}=Msg 
+ 	when OtherNode /= node() ->
+ 	    {noreply,NS} = handle_info(Msg,S),
+ 	    dist_process_requests(NS,Timeout)
+ 
+     after Timeout ->
+ 	    S
+     end.
+ 
  restart_appl(AppName, S) ->
***************
*** 947,949 ****
  	{value, Appl} when element(1, Appl#appl.id) == failover ->
! 	    case catch start_appl(AppName, S, req) of
  		{ok, NewS, _} ->
--- 1129,1131 ----
  	{value, Appl} when element(1, Appl#appl.id) == failover ->
! 	    case catch locked_start_appl(AppName, S, req) of
  		{ok, NewS, _} ->
***************
*** 984,986 ****
  permit(true, _, AppName, From, S, LockId) ->
!     case catch start_appl(AppName, S, req) of
  	{_ErrorTag, {not_running, App}} ->
--- 1166,1168 ----
  permit(true, _, AppName, From, S, LockId) ->
!     case catch locked_start_appl(AppName, S, req) of
  	{_ErrorTag, {not_running, App}} ->
***************
*** 1008,1010 ****
        fun(AppName, NewS) ->
! 	      case catch start_appl(AppName, NewS, req) of
  		  {error, R}  ->
--- 1190,1192 ----
        fun(AppName, NewS) ->
! 	      case catch locked_start_appl(AppName, NewS, req) of
  		  {error, R}  ->
***************
*** 1034,1040 ****
  		{value, {Node, _Name}} ->
! 		    {already_started, Node};
  		_ when AllNodes /= [] ->
! 		    not_started;
  		_ ->
! 		    no_permission
  	    end
--- 1216,1222 ----
  		{value, {Node, _Name}} ->
! 		    {{already_started, Node},S};
  		_ when AllNodes /= [] ->
! 		    {not_started,S};
  		_ ->
! 		    {no_permission,S}
  	    end
***************
*** 1045,1047 ****
      case find_any_node(tuple_to_list(AnyNodes), Name, S, Weight, AllNodes) of
! 	false -> find_start_node(Nodes, Name, S, Weight, AllNodes);
  	Res -> Res
--- 1227,1230 ----
      case find_any_node(tuple_to_list(AnyNodes), Name, S, Weight, AllNodes) of
! 	{false,NS} -> 
! 	    find_start_node(Nodes, Name, NS, Weight, AllNodes);
  	Res -> Res
***************
*** 1053,1055 ****
  		{value, #appl{id = {distributed, Node}}} ->
! 		    {already_started, Node};
  		_ ->
--- 1236,1238 ----
  		{value, #appl{id = {distributed, Node}}} ->
! 		    {{already_started, Node},S};
  		_ ->
***************
*** 1057,1061 ****
  			{value, {Node, _Name}} ->
! 			    {already_started, Node};
  			_ ->
! 			    {ok, Node}
  		    end
--- 1240,1244 ----
  			{value, {Node, _Name}} ->
! 			    {{already_started, Node},S};
  			_ ->
! 			    {{ok, Node},S}
  		    end
***************
*** 1064,1067 ****
      end;
! find_start_node([], _Name, _S, _Weight, _AllNodes) ->
!     not_started.
  
--- 1247,1250 ----
      end;
! find_start_node([], _Name, S, _Weight, _AllNodes) ->
!     {not_started,S}.
  
***************
*** 1073,1075 ****
      case check_running(Name, S, intersection(AnyNodes, AllNodes)) of
! 	{already_started, Node} -> {already_started, Node};
  	false ->
--- 1256,1260 ----
      case check_running(Name, S, intersection(AnyNodes, AllNodes)) of
! 	{already_started, Node} -> 
! 	    {{already_started, Node},S};
! 
  	false ->
***************
*** 1077,1083 ****
  	    send_nodes(AllNodes, {dist_ac_weight, Name, Weight, node()}),
! 	    Answers = [{Weight, node()} |
! 		       collect_answers(AllNodes, Name, S, [])],
  	    %% Make a decision (the same at every node) (smallest weight wins)
! 	    find_alive_node(lists:sort(Answers), 
! 			    intersection(AnyNodes, S#state.known))
      end.
--- 1262,1268 ----
  	    send_nodes(AllNodes, {dist_ac_weight, Name, Weight, node()}),
! 	    {CollectedAnswers,NS} = collect_answers(AllNodes, Name, S, []),
! 	    Answers = [{Weight, node()} | CollectedAnswers ],
  	    %% Make a decision (the same at every node) (smallest weight wins)
! 	    {find_alive_node(lists:sort(Answers), 
! 			     intersection(AnyNodes, NS#state.known)),NS}
      end.
***************
*** 1108,1109 ****
--- 1293,1296 ----
  
+ find_alive_node([{undefined, Node} | Nodes], AliveNodes) ->
+     find_alive_node(Nodes, AliveNodes);
  find_alive_node([{_, Node} | Nodes], AliveNodes) ->
***************
*** 1131,1135 ****
  	    receive
! 		{dist_ac_weight, Name, Weight, Node} ->
  		    monitor_node(Node, false),
  		    collect_answers(Nodes, Name, S, [{Weight, Node} | Res]);
  		{nodedown, Node} ->
--- 1318,1338 ----
  	    receive
! 		{dist_ac_new_node, Vsn, OtherNode, HisAppls, HisStarted}=Msg ->
! 		    NS = handle_dist_request(Msg,S),
! 		    monitor_node(Node, false),
! 		    collect_answers([Node|Nodes], Name, NS, Res);
! 
! 		{dist_ac_app_state, OtherNode, OtherAppName}=Msg ->
! 		    NS = handle_dist_request(Msg,S),
! 		    monitor_node(Node, false),
! 		    collect_answers([Node|Nodes], Name, NS, Res);
! 
! 		{dist_ac_weight, Name, Weight, OtherNode}=Msg when OtherNode /= node() ->
! 		    NS = handle_dist_request(Msg,S),
! 		    monitor_node(Node, false),
! 		    collect_answers([Node|Nodes], Name, NS, Res);
! 
! 		{dist_ac_weight_resp, Name, Weight, Node} ->
  		    monitor_node(Node, false),
  		    collect_answers(Nodes, Name, S, [{Weight, Node} | Res]);
+ 
  		{nodedown, Node} ->
***************
*** 1141,1144 ****
      collect_answers(Nodes, Name, S, Res);
! collect_answers([], _Name, _S, Res) ->
!     Res.
  
--- 1344,1347 ----
      collect_answers(Nodes, Name, S, Res);
! collect_answers([], _Name, S, Res) ->
!     {Res,S}.
  
***************
*** 1514,1517 ****
--- 1717,1722 ----
  
+ -ifdef(UNUSED_FUNCTIONS).
  error_msg(Format) when list(Format) ->
      error_msg(Format, []).
+ -endif.
  
***************
*** 1520,1521 ****
--- 1725,1727 ----
  
+ -ifdef(UNUSED_FUNCTIONS).
  info_msg(Format) when list(Format) ->
***************
*** 1525 ****
--- 1731,1844 ----
      error_logger:info_msg("dist_ac on node ~p:~n" ++ Format, [node()|ArgList]).
+ -endif.
+ 
+ %% some helpers
+ 
+ dist_ac_version_mismatch(Vsn, Node) ->
+     error_msg("Node ~p protocol version ~p is incompatible with version ~p. Terminating node ~p.\n",
+ 	      [Node,Vsn,?vsn,node()]),
+     init:stop(),
+     KernelSup = whereis(kernel_sup),
+     receive
+ 	{'EXIT',KernelSup,Reason} = X ->
+ 	    exit(Reason)
+     end.
+ 
+ handle_dist_request({dist_ac_new_node,Vsn,Node,HisAppls,_HisStarted},S) when Vsn == 1 ->
+     dist_ac_version_mismatch(Vsn, Node);
+ 
+ handle_dist_request({dist_ac_new_node,Vsn,Node,HisAppls,_HisStarted},S) ->
+     Appls = S#state.appls,
+     MyStarted = zf(fun(Appl) when Appl#appl.id == local ->
+ 			   {true, {node(), Appl#appl.name}};
+ 		      (_) ->
+ 			   false
+ 		   end, Appls),
+     {?DIST_AC, Node} ! {dist_ac_new_node_resp, ?vsn, node(), Appls, MyStarted},
+     NAppls = dist_merge(Appls, HisAppls, Node),
+     S#state{appls = NAppls, known = [Node | S#state.known]};
+ 
+ handle_dist_request({dist_ac_weight, Name, Weight, Node}, S) ->
+     %% This means another node starts up, and will eventually take over
+     %% this appl.  We have a situation like: {Name, [{Node}, node()]}
+     %% Node sends us this msg, and we must respond.  It doesn't really
+     %% matter what we send him; but it must be a dist_ac_weight msg.
+     %% Another situation is {Name, [RNode, {node()}, Node]}.
+     %%
+     %% Yet another situation is that the node where Name was running crashed,
+     %% and Node has got the nodedown message, but we haven't.  In this case,
+     %% we must send a correct weight to Node. i.e. the same weight that
+     %% we'll send to him later, when we get the nodedown message.
+     case keysearch(Name, #appl.name, S#state.appls) of
+ 	{value, Appl} -> 
+ 	    Id = Appl#appl.id,
+ 	    case Id of 
+ 		run_waiting ->
+ 		    {?DIST_AC, Node} ! {dist_ac_weight_resp, Name, 0, node()},
+ 		   S;
+ 		undefined -> 
+ 		    {?DIST_AC, Node} ! {dist_ac_weight_resp, Name, undefined, node()},
+ 		    S#state{tmp_locals = [{Name, Weight, Node} |
+ 					  S#state.tmp_locals]};
+ 		{takeover, _} -> 
+ 		    {?DIST_AC, Node} ! {dist_ac_weight_resp, Name, undefined, node()},
+ 		    S#state{tmp_locals = [{Name, Weight, Node} |
+ 					  S#state.tmp_locals]};
+ 		{failover, _} -> 
+ 		    {?DIST_AC, Node} ! {dist_ac_weight_resp, Name, undefined, node()},
+ 		    S#state{tmp_locals = [{Name, Weight, Node} |
+ 					  S#state.tmp_locals]};
+ 		_ ->
+ 		    MyWeight = get_cached_weight(Name, S),
+ 		    {?DIST_AC, Node} ! {dist_ac_weight_resp, Name, MyWeight, node()},
+ 		    NTWs = keyreplaceadd(Name, 1, S#state.tmp_weights,
+ 					 {Name, MyWeight}),
+ 		    S#state{tmp_weights = NTWs}
+ 	    end;
+ 	_ ->
+ 	    {?DIST_AC, Node} ! {dist_ac_weight_resp, Name, undefined, node()},
+ 	    S#state{tmp_locals = [{Name, Weight, Node} | S#state.tmp_locals]}
+     end;
+ 
+ handle_dist_request({dist_ac_app_state, Node, AppName}=Msg, S) ->
+     handle_dist_request({Msg,false}, S);
+ 
+ handle_dist_request({{dist_ac_app_state, Node, AppName},BlockEqual}, S) ->
+     AppNames = dist_get_runnable(S#state.appls),
+     {AppId,DontStart} = 
+ 	case {lists:member(AppName,AppNames),
+ 	      lists:keysearch(AppName, #appl.name, S#state.appls)} of
+ 	    {true,{value, Appl}} when Appl#appl.id =/= undefined ->
+ 		%% So we have this application running ...
+ 		%% check if we should hint the other side not to start
+ 		TmpDontStart =
+ 		    case {node_have_priority(node(),Node,Appl#appl.nodes),BlockEqual} of
+ 			{equal,true} ->
+ 			    true;
+ 			{equal,_} ->
+ 			    false;
+ 			{TmpPrio,_} ->
+ 			    TmpPrio
+ 		    end,
+ 		{Appl#appl.id, TmpDontStart};
+ 	    _ ->
+ 		{undefined,false}
+ 	end,
+     {?DIST_AC, Node} ! { dist_ac_app_state_resp, node(), {AppName, AppId, DontStart}},
+     S.
+ 
+ node_have_priority(_,_,[]) ->
+     false;
+ node_have_priority(N1,N2,[TupleNodes|T]) when tuple(TupleNodes) ->
+     Nodes = tuple_to_list(TupleNodes),
+     case {lists:member(N1,Nodes),lists:member(N2,Nodes)} of
+ 	{true,true} -> equal;
+ 	{true,false} -> true;
+ 	{false,true} -> false;
+ 	_ ->
+ 	    node_have_priority(N1,N2,T)
+     end;
+ node_have_priority(N1,N2,[N1|Nodes]) -> true;
+ node_have_priority(N1,N2,[N2|Nodes]) -> false;
+ node_have_priority(N1,N2,[NX|Nodes]) ->
+     node_have_priority(N1,N2,Nodes).
+     
*** lib/kernel/src/application_controller.erl	Mon Jan 28 09:41:23 2002
--- ../otp_src_R8B-1.fixed/lib/kernel/src/application_controller.erl	Mon May  6 23:48:38 2002
***************
*** 35,37 ****
  %% Internal exports
! -export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2,
  	 init_starter/4, get_loaded/1]).
--- 35,38 ----
  %% Internal exports
! -export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2, 
! 	 code_change/3,
  	 init_starter/4, get_loaded/1]).
***************
*** 1030,1031 ****
--- 1031,1036 ----
  		    stop_appl(AppName, Id, Type),
+ 		    case Id of 
+ 			{distributed,Node} -> ignore ;
+ 			_ -> info_started(AppName,Node)
+ 		    end,
  		    NRunning = [{AppName, {distributed, Node}} |
***************
*** 1103,1104 ****
--- 1108,1118 ----
  		{started, Node} ->
+ 		    case keysearch(AppName,1,Running) of
+ 			{value, {AppName,Id}} ->
+ 			    case Id of
+ 				{distributed, Node} -> ignore;
+ 				_ -> info_started(AppName,Node)
+ 			    end;
+ 			_ ->
+ 			    info_started(AppName,Node)
+ 		    end,
  		    NRunning = [{AppName, {distributed, Node}} |
***************
*** 1151,1152 ****
--- 1165,1169 ----
  
+ code_change(OldVsn,S,Extra) ->
+     {ok,S}.
+ 
  %%%-----------------------------------------------------------------
***************
*** 1326,1327 ****
--- 1343,1345 ----
  %%-----------------------------------------------------------------
+ 
  stop_appl(AppName, Id, Type) when pid(Id) ->
*** lib/kernel/src/application.erl	Mon Oct  8 10:29:53 2001
--- ../otp_src_R8B-1.fixed/lib/kernel/src/application.erl	Wed Apr 24 12:27:03 2002
***************
*** 27,28 ****
--- 27,29 ----
  -export([start_type/0]).
+ -export([whereis_application/1]).
  
***************
*** 142,143 ****
--- 143,146 ----
  
+ whereis_application(Application) ->
+     dist_ac:whereis_application(Application).
  
